/*! \file cfrfme.cpp
 *
 * C++ implementation of FRFME functions.
 */
#include <complex>
#include <cstdio>
#include <exception>
#include <fstream>
#include <regex>
#include <string>

#ifndef INCLUDE_PARSERS_H_
#include "../include/Parsers.h"
#endif

#ifndef INCLUDE_COMMONS_H_
#include "../include/Commons.h"
#endif

#ifndef INCLUDE_CONFIGURATION_H_
#include "../include/Configuration.h"
#endif

#ifndef INCLUDE_SPH_SUBS_H_
#include "../include/sph_subs.h"
#endif

#ifndef INCLUDE_TFRFME_H_
#include "../include/tfrfme.h"
#endif

#ifndef INCLUDE_TRA_SUBS_H_
#include "../include/tra_subs.h"
#endif

using namespace std;

/*! \brief C++ implementation of FRFME
 *
 *  \param data_file: `string` Name of the input data file.
 *  \param output_path: `string` Directory to write the output files in.
 */
void frfme(string data_file, string output_path) {
  string tfrfme_name = output_path + "/c_TFRFME.hd5";
  //fstream tfrfme;
  TFRFME *tfrfme = NULL;
  char namef[7];
  char more;
  double *vkv = NULL, **vkzm = NULL;
  complex<double> *wk = NULL, **w = NULL;
  const complex<double> cc0(0.0, 0.0);
  const complex<double> uim(0.0, 1.0);
  int line_count = 0, last_read_line = 0;
  regex re = regex("-?[0-9]+");
  string *file_lines = load_file(data_file, &line_count);
  smatch m;
  string str_target = file_lines[last_read_line++];
  regex_search(str_target, m, re);
  int jlmf = stoi(m.str());
  str_target = m.suffix().str();
  regex_search(str_target, m, re);
  int jlml = stoi(m.str());
  int lmode = 0, lm = 0, nks = 0, nkv = 0;
  double vk = 0.0, exri = 0.0, an = 0.0, ff = 0.0, tra = 0.0;
  double exdc = 0.0, wp = 0.0, xip = 0.0, xi = 0.0;
  int idfc = 0, nxi = 0;
  double apfafa = 0.0, pmf = 0.0, spd = 0.0, rir = 0.0, ftcn = 0.0, fshmx = 0.0;
  double vxyzmx = 0.0, delxyz = 0.0, vknmx = 0.0, delk = 0.0, delks = 0.0;
  double frsh = 0.0, exril = 0.0;
  int nlmmt = 0, nrvc = 0;
  // Vector size variables
  int wsum_size;
  // End of vector size variables
  if (jlmf != 1) {
    int nxv, nyv, nzv;
    if (tfrfme == NULL) tfrfme = TFRFME::from_binary(tfrfme_name, "HDF5");
    if (tfrfme != NULL) {
      lmode = (int)tfrfme->get_param("lmode");
      lm = (int)tfrfme->get_param("lm");
      nkv = (int)tfrfme->get_param("nkv");
      nxv = (int)tfrfme->get_param("nkv");
      nyv = (int)tfrfme->get_param("nkv");
      nzv = (int)tfrfme->get_param("nkv");
      vk = tfrfme->get_param("vk");
      exri = tfrfme->get_param("exri");
      an = tfrfme->get_param("an");
      ff = tfrfme->get_param("ff");
      tra = tfrfme->get_param("tra");
      spd = tfrfme->get_param("spd");
      frsh = tfrfme->get_param("frsh");
      exril = tfrfme->get_param("exril");
      vkv = new double[nkv]();
      fstream temptape2;
      string tempname2 = output_path + "/c_TEMPTAPE2";
      temptape2.open(tempname2.c_str(), ios::in | ios::binary);
      if (temptape2.is_open()) {
	for (int jx = 0; jx < nkv; jx++) temptape2.read(reinterpret_cast<char *>(&(vkv[jx])), sizeof(double));
	vkzm = new double*[nkv];
	for (int vki = 0; vki < nkv; vki++) vkzm[vki] = new double[nkv]();
	for (int jy10 = 0; jy10 < nkv; jy10++) {
	  for (int jx10 = 0; jx10 < nkv; jx10++) {
	    temptape2.read(reinterpret_cast<char *>(&(vkzm[jx10][jy10])), sizeof(double));
	  } //jx10 loop
	} // jy10 loop
	temptape2.read(reinterpret_cast<char *>(&apfafa), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&pmf), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&spd), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&rir), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&ftcn), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&fshmx), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&vxyzmx), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&delxyz), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&vknmx), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&delk), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&delks), sizeof(double));
	temptape2.read(reinterpret_cast<char *>(&nlmmt), sizeof(int));
	temptape2.read(reinterpret_cast<char *>(&nrvc), sizeof(int));
	temptape2.close();
      } else {
	printf("ERROR: could not open TEMPTAPE2 file.\n");
      }
    } else {
      printf("ERROR: could not open TFRFME file.\n");
    }
    nks = nkv - 1;
  } else { // label 16
    int nksh, nrsh, nxsh, nysh, nzsh;
    str_target = file_lines[last_read_line++];
    for (int cli = 0; cli < 7; cli++) {
      regex_search(str_target, m, re);
      if (cli == 0) lmode = stoi(m.str());
      else if (cli == 1) lm = stoi(m.str());
      else if (cli == 2) nksh = stoi(m.str());
      else if (cli == 3) nrsh = stoi(m.str());
      else if (cli == 4) nxsh = stoi(m.str());
      else if (cli == 5) nysh = stoi(m.str());
      else if (cli == 6) nzsh = stoi(m.str());
      str_target = m.suffix().str();
    }
    re = regex("-?[0-9]\\.[0-9]+([dDeE][-+]?[0-9]+)?");
    regex_search(str_target, m, re);
    double wlenfr = stod(m.str());
    str_target = file_lines[last_read_line++];
    for (int cli = 0; cli < 3; cli++) {
      regex_search(str_target, m, re);
      if (cli == 0) an = stod(m.str());
      else if (cli == 1) ff = stod(m.str());
      else if (cli == 2) tra = stod(m.str());
      str_target = m.suffix().str();
    }
    double spdfr, exdcl;
    str_target = file_lines[last_read_line++];
    for (int cli = 0; cli < 3; cli++) {
      regex_search(str_target, m, re);
      if (cli == 0) spd = stod(m.str());
      else if (cli == 1) spdfr = stod(m.str());
      else if (cli == 2) exdcl = stod(m.str());
      str_target = m.suffix().str();
    }
    str_target = file_lines[last_read_line++];
    re = regex("[eEmM]");
    if (regex_search(str_target, m, re)) {
      more = m.str().at(0);
      if (more == 'm' || more == 'M') {
	more = 'M';
	sprintf(namef, "c_TMDF");
      }
      else if (more == 'e' || more == 'E') {
	more = 'E';
	sprintf(namef, "c_TEDF");
      }
      str_target = m.suffix().str();
      re = regex("[0-9]+");
      regex_search(str_target, m, re);
      int ixi = stoi(m.str());
      string tedf_name = output_path + "/" + namef + ".hd5";
      ScattererConfiguration *tedf = ScattererConfiguration::from_binary(tedf_name, "HDF5");
      if (tedf != NULL) {
	int iduml, idum;
	iduml = (int)tedf->get_param("nsph");
	idum = tedf->get_iog(iduml - 1);
	exdc = tedf->get_param("exdc");
	wp = tedf->get_param("wp");
	xip = tedf->get_param("xip");
	idfc = (int)tedf->get_param("idfc");
	nxi = (int)tedf->get_param("nxi");
	if (idfc >= 0) {
	  if (ixi <= nxi) {
	    xi = tedf->get_scale(ixi - 1);
	  } else { // label 96
	    delete tedf;
	    // label 98
	    string output_name = output_path + "/c_OFRFME";
	    FILE *output = fopen(output_name.c_str(), "w");
	    fprintf(output, "  WRONG INPUT TAPE\n");
	    fclose(output);
	  }
	} else { // label 18
	  xi = xip;
	}
	// label 20
	delete tedf;
	double wn = wp / 3.0e8;
	vk = xi * wn;
	exri = sqrt(exdc);
	frsh = 0.0;
	exril = 0.0;
	fshmx = 0.0;
	apfafa = exri / (an * ff);
	if (lmode != 0) pmf = 2.0 * apfafa;
	if (spd > 0.0) {
	  exril = sqrt(exdcl);
	  rir = exri / exril;
	  ftcn = 2.0 / (1.0 + rir);
	  frsh = -spd * spdfr;
	  double sthmx = an / exri;
	  double sthlmx = sthmx * rir;
	  double uy = 1.0;
	  fshmx = spd * (rir * (sqrt(uy - sthmx * sthmx) / sqrt(uy - sthlmx * sthlmx)) - uy);
	}
	// label 22
	nlmmt = lm * (lm + 2) * 2;
	nks = nksh * 2;
	nkv = nks + 1;
	// Array initialization
	vkv = new double[nkv]();
	vkzm = new double*[nkv];
	for (int vi = 0; vi < nkv; vi++) vkzm[vi] = new double[nkv];
	// End of array initialization
	double vkm = vk * exri;
	vknmx = vk * an;
	delk = vknmx / nksh;
	delks = delk / vkm;
	delks = delks * delks;
	vxyzmx = acos(0.0) * 4.0 / vkm * wlenfr;
	delxyz = vxyzmx / nrsh;
	int nxs = nxsh * 2;
	int nxv = nxs + 1;
	int nxshpo = nxsh + 1;
	int nys = nysh * 2;
	int nyv = nys + 1;
	int nyshpo = nysh + 1;
	int nzs = nzsh * 2;
	int nzv = nzs + 1;
	int nzshpo = nzsh + 1;
	tfrfme = new TFRFME(lmode, lm, nkv, nxv, nyv, nzv);
	for (int i24 = nxshpo; i24 <= nxs; i24++) {
	  tfrfme->set_x(i24, tfrfme->get_x(i24 - 1) + delxyz);
	  tfrfme->set_x(nxv - i24 - 1, -tfrfme->get_x(i24));
	} // i24 loop
	for (int i25 = nyshpo; i25 <= nys; i25++) {
	  tfrfme->set_y(i25, tfrfme->get_y(i25 - 1) + delxyz);
	  tfrfme->set_y(nyv - i25 - 1, -tfrfme->get_y(i25));
	} // i25 loop
	for (int i27 = nzshpo; i27 <= nzs; i27++) {
	  tfrfme->set_z(i27, tfrfme->get_z(i27 - 1) + delxyz);
	  tfrfme->set_z(nzv - i27 - 1, -tfrfme->get_z(i27));
	} // i27 loop
	int nrvc = nxv * nyv * nzv;
	int nkshpo = nksh + 1;
	for (int i28 = nkshpo; i28 <= nks; i28++) {
	  vkv[i28] = vkv[i28 - 1] + delk;
	  vkv[nkv - i28 - 1] = -vkv[i28];
	} // i28 loop
	if (tfrfme != NULL) {
	  tfrfme->set_param("vk", vk);
	  tfrfme->set_param("exri", exri);
	  tfrfme->set_param("an", an);
	  tfrfme->set_param("ff", ff);
	  tfrfme->set_param("tra", tra);
	  tfrfme->set_param("spd", spd);
	  tfrfme->set_param("frsh", frsh);
	  tfrfme->set_param("exril", exril);
	  fstream temptape1, temptape2;
	  string temp_name1 = output_path + "/c_TEMPTAPE1";
	  string temp_name2 = output_path + "/c_TEMPTAPE2";
	  temptape1.open(temp_name1.c_str(), ios::out | ios::binary);
	  temptape2.open(temp_name2.c_str(), ios::out | ios::binary);
	  for (int jx = 0; jx < nkv; jx++)
	    temptape2.write(reinterpret_cast<char *>(&(vkv[jx])), sizeof(double));
	  frfmer(nkv, vkm, vkv, vknmx, apfafa, tra, spd, rir, ftcn, lm, lmode, pmf, temptape1, temptape2);
	  temptape1.close();
	  temptape2.write(reinterpret_cast<char *>(&apfafa), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&pmf), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&spd), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&rir), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&ftcn), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&fshmx), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&vxyzmx), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&delxyz), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&vknmx), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&delk), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&delks), sizeof(double));
	  temptape2.write(reinterpret_cast<char *>(&nlmmt), sizeof(int));
	  temptape2.write(reinterpret_cast<char *>(&nrvc), sizeof(int));
	  temptape2.close();
	  temptape2.open("c_TEMPTAPE2", ios::in | ios::binary);
	  for (int jx = 0; jx < nkv; jx++) {
	    double value = 0.0;
	    temptape2.read(reinterpret_cast<char *>(&value), sizeof(double));
	    vkv[jx] = value;
	  }
	  for (int jy40 = 0; jy40 < nkv; jy40++) {
	    for (int jx40 = 0; jx40 < nkv; jx40++) {
	      double value = 0.0;
	      temptape2.read(reinterpret_cast<char *>(&value), sizeof(double));
	      vkzm[jx40][jy40] = value;
	    }
	  } // jy40 loop
	  temptape2.close();
	  for (int j80 = jlmf - 1; j80 < jlml; j80++) {
	    // w matrix
	    if (w != NULL) {
	      for (int wi = nkv - 1; wi > -1; wi--) delete[] w[wi];
	      delete[] w;
	    }
	    w = new complex<double>*[nkv];
	    for (int wi = 0; wi < nkv; wi++) w[wi] = new complex<double>[nkv]();
	    if (wk != NULL) delete[] wk;
	    wk = new complex<double>[nlmmt]();
	    temptape1.open(temp_name1.c_str(), ios::in | ios::binary);
	    for (int jy50 = 0; jy50 < nkv; jy50++) {
	      for (int jx50 = 0; jx50 < nkv; jx50++) {
		for (int i = 0; i < nlmmt; i++) {
		  double vreal, vimag;
		  temptape1.read(reinterpret_cast<char *>(&vreal), sizeof(double));
		  temptape1.read(reinterpret_cast<char *>(&vimag), sizeof(double));
		  wk[i] = complex<double>(vreal, vimag);
		}
		w[jx50][jy50] = wk[j80];
	      } // jx50
	    } // jy50 loop
	    temptape1.close();
	    int ixyz = 0;
	    for (int wj = 0; wj < nrvc; wj++) tfrfme->set_matrix_element(j80, wj, cc0);
	    for (int iz75 = 0; iz75 < nzv; iz75++) {
	      double z = tfrfme->get_z(iz75)  + frsh;
	      for (int iy70 = 0; iy70 < nyv; iy70++) {
		double y = tfrfme->get_y(iy70);
		for (int ix65 = 0; ix65 < nxv; ix65++) {
		  double x = tfrfme->get_x(ix65);
		  ixyz++;
		  complex<double> sumy = cc0;
		  for (int jy60 = 0; jy60 < nkv; jy60++) {
		    double vky = vkv[jy60];
		    double vkx = vkv[nkv - 1];
		    double vkzf = vkzm[0][jy60];
		    complex<double> phasf = exp(uim * (-vkx * x + vky * y +vkzf * z));
		    double vkzl = vkzm[nkv - 1][jy60];
		    complex<double> phasl = exp(uim * (vkx * x + vky * y + vkzl * z));
		    complex<double> sumx = 0.5 * (w[0][jy60] * phasf + w[nkv - 1][jy60] * phasl);
		    for (int jx55 = 2; jx55 <= nks; jx55++) {
		      vkx = vkv[jx55 - 1];
		      double vkz = vkzm[jx55 - 1][jy60];
		      complex<double> phas = exp(uim * (vkx * x + vky * y + vkz * z));
		      sumx += (w[jx55 - 1][jy60] * phas);
		    } // jx55 loop
		    if (jy60 == 0 || jy60 == nkv - 1) sumx *= 0.5;
		    sumy += sumx;
		  } // jy60 loop
		  tfrfme->set_matrix_element(j80, ixyz - 1, sumy * delks);
		} // ix65 loop
	      } // iy70 loop
	    } // iz75 loop
	  } // j80 loop
	  if (jlmf != 1) {
	    lmode = (int)tfrfme->get_param("lmode");
	    lm = (int)tfrfme->get_param("lm");
	    nkv = (int)tfrfme->get_param("nkv");
	    nxv = (int)tfrfme->get_param("nxv");
	    nyv = (int)tfrfme->get_param("nyv");
	    nzv = (int)tfrfme->get_param("nzv");
	    vk = tfrfme->get_param("vk");
	    exri = tfrfme->get_param("exri");
	    an = tfrfme->get_param("an");
	    ff = tfrfme->get_param("ff");
	    tra = tfrfme->get_param("tra");
	    spd = tfrfme->get_param("spd");
	    frsh = tfrfme->get_param("frsh");
	    exril = tfrfme->get_param("exril");
	  }
	  // label 88
	  tfrfme->write_binary(tfrfme_name, "HDF5");
	  string output_name = output_path + "/c_OFRFME";
	  FILE *output = fopen(output_name.c_str(), "w");
	  fprintf(output, " IF JLML < NLMMT, PRESERVE TEMPTAPE1, TEMPTAPE2, AND TFRFRME,\n");
	  fprintf(output, " AND RESTART LM RUN WITH JLMF = JLML+1\n");
	  if (spd > 0.0) fprintf(output, "  FSHMX =%15.7lE\n", fshmx);
	  fprintf(output, "  FRSH =%15.7lE\n", frsh);
	  fclose(output);
	} else { // Should never happen.
	  printf("ERROR: could not open TFRFME file for output.\n");
	}
      } else {
	printf("ERROR: could not open TEDF file.\n");
      }
    } else { // label 98
      string output_name = output_path + "/c_OFRFME";
      FILE *output = fopen(output_name.c_str(), "w");
      fprintf(output, "  WRONG INPUT TAPE\n");
      fclose(output);
    }
  }
  // label 45
  if (tfrfme != NULL) delete tfrfme;
  delete[] file_lines;
  if (vkv != NULL) delete[] vkv;
  if (vkzm != NULL) {
    for (int vki = nkv - 1; vki > -1; vki--) delete[] vkzm[vki];
    delete[] vkzm;
  }
  if (w != NULL) {
    for (int wi = nkv - 1; wi > -1; wi--) delete[] w[wi];
    delete[] w;
  }
  if (wk != NULL) delete[] wk;
  printf("Done.\n");
}
